#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 18:12
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
from xmlrpc.client import Fault

import numpy as np
from ase import Atoms
from ase.io import write as ase_write
from rich.progress import track

from NepTrain import utils
from NepTrain.core.select import filter_by_bonds


def perturb_position(prim,min_distance):

    atoms = prim.copy()
    # 获取原子位置
    positions = atoms.get_positions()

    # 添加随机微扰
    perturbed_positions = positions + np.random.uniform(
        low=-min_distance,
        high=min_distance,
        size=positions.shape
    )

    # 更新结构的原子位置
    atoms.set_positions(perturbed_positions)
    return atoms

def generate_strained_structure(prim, strain_lim,min_distance):
    strains = np.random.uniform(*strain_lim, (3, ))
    atoms = prim.copy()
    cell_new = prim.cell[:] * (1 + strains)
    atoms.set_cell(cell_new, scale_atoms=True)

    return perturb_position(atoms,min_distance)


def generate_deformed_structure(prim, strain_lim,min_distance):
    R = np.random.uniform(*strain_lim, (3, 3))
    M = np.eye(3) + R
    atoms = prim.copy()
    cell_new = M @ atoms.cell[:]
    atoms.set_cell(cell_new, scale_atoms=True)

    return perturb_position(atoms,min_distance)



@utils.iter_path_to_atoms(["*.vasp","*.xyz"],show_progress=False )
def perturb(atoms:Atoms,cell_pert_fraction=0.04, min_distance=0.1,num=50,filter_bonds=False):
    cell_pert_fraction=[-cell_pert_fraction,cell_pert_fraction]
    structures_rattle=[]
    for i in track(range(num),description=f"Current structure:{atoms.symbols}"):
        if i%2==0:
            structure=generate_deformed_structure(atoms,cell_pert_fraction,min_distance)
            structure.info['Config_type'] = f"perturb {i+1} deformed {cell_pert_fraction}  min_distance {min_distance}"

        else:
            structure=generate_strained_structure(atoms,cell_pert_fraction,min_distance)
            structure.info['Config_type'] = f"perturb {i+1} strained {cell_pert_fraction}  min_distance {min_distance}"

        structures_rattle.append(structure)
    if filter_bonds:
        structures_rattle, bad = filter_by_bonds(structures_rattle, model=atoms)


    return structures_rattle
def run_perturb(argparse):

    result = perturb(argparse.model_path,
                     cell_pert_fraction=argparse.cell_pert_fraction,
                     min_distance=argparse.min_distance,

                     num=argparse.num,
                     filter_bonds=argparse.filter
                     )
    path=os.path.dirname(argparse.out_file_path)

    if path and  not os.path.exists(path):
        os.makedirs(path)

    ase_write(argparse.out_file_path,[atom for _list in result for atom in _list],format="extxyz",append=argparse.append)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 19:03
# @Author  : 兵
# @email    : 1747193328@qq.com

import math
import os.path

import numpy as np
from ase import Atoms
from ase.io import write as ase_write

from NepTrain import utils, Config, module_path
from ..utils import check_env
from .io import VaspInput

atoms_index=1

@utils.iter_path_to_atoms(["*.vasp","*.xyz"],show_progress=True,
                 description="VASP calculation progress" )
def calculate_vasp(atoms:Atoms,argparse):
    global atoms_index

    vasp = VaspInput()
    if argparse.incar is not None and os.path.exists(argparse.incar):
        vasp.read_incar(argparse.incar)
    else:
        vasp.read_incar(os.path.join(module_path,"core/vasp/INCAR"))
    directory=os.path.join(argparse.directory,f"{atoms_index}-{atoms.symbols}")
    atoms_index+=1
    command=f"{Config.get('environ','mpirun_path')} -n {argparse.n_cpu} {Config.get('environ','vasp_path')}"

    a,b,c,alpha, beta, gamma=atoms.get_cell_lengths_and_angles()
    if argparse.kspacing is not None:
        vasp.set(kspacing=argparse.kspacing)
    vasp.set(
             directory=directory,
             command=command,
            kpts=(math.ceil(argparse.ka[0]/a)  ,
                  math.ceil(argparse.ka[1]/b)  ,
                  math.ceil(argparse.ka[2]/b) ),
             gamma=argparse.use_gamma,
             )

    vasp.calculate(atoms,('energy'))
    atoms.calc=vasp._xml_calc
    xx, yy, zz, yz, xz, xy = -vasp.results['stress'] * atoms.get_volume()  # *160.21766
    atoms.info['virial'] = np.array([(xx, xy, xz), (xy, yy, yz), (xz, yz, zz)])
    #这里没想好怎么设计config的格式化  就先使用原来的
    if "Config_type" not in atoms.info:

        atoms.info['Config_type'] = "NepTrain scf "
    atoms.info['Weight'] = 1.0
    del atoms.calc.results['stress']
    del atoms.calc.results['free_energy']


    if vasp.converged:
        return atoms
    else:
        raise ValueError(f"{directory}: VASP not converged")
def run_vasp(argparse):
    check_env()

    result = calculate_vasp(argparse.model_path,argparse)
    path=os.path.dirname(argparse.out_file_path)
    if path and  not os.path.exists(path):
        os.makedirs(path)



    ase_write(argparse.out_file_path,result,format="extxyz",append=argparse.append)

    utils.print_success("VASP calculation task completed!" )


if __name__ == '__main__':
    calculate_vasp("./")


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 16:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import configparser
import os
import shutil

from watchdog.observers import Observer

from NepTrain import utils

__version__="1.4.3"



config_path = utils.get_config_path()

module_path = os.path.dirname(__file__)

if not os.path.exists(config_path)  :
    shutil.copy(os.path.join(module_path,"config.ini"), config_path)

Config = configparser.RawConfigParser()
Config.read(config_path,encoding="utf8")



#
# if platform.is_linux():
#     # wsl测试默认的有问题  强行切换到poll机制
#     observer = PollingObserver()
# else:
#     observer = Observer()


observer = Observer()



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 19:52
# @Author  : 兵
# @email    : 1747193328@qq.com



from scipy.spatial.distance import cdist


from NepTrain.core.nep.utils import get_descriptor_function


def farthest_point_sampling(points, n_samples, min_dist=0.1, selected_data=None):
    """
    最远点采样：支持已有样本扩展，并加入最小距离限制。

    参数:
        points (ndarray): 点集，形状为 (N, D)。
        n_samples (int): 最大采样点的数量。
        min_dist (float): 最小距离阈值。
        initial_indices (list or None): 已选择的样本索引列表，默认无。

    返回:
        sampled_indices (list): 采样点的索引。
    """
    n_points = points.shape[0]

    if isinstance(selected_data, np.ndarray) and selected_data.size == 0:
        selected_data=None
    # 初始化采样点列表
    sampled_indices = []

    # 如果已有采样点，则计算到所有点的最小距离
    if selected_data is not None :
        # 使用 cdist 计算已有点与所有点之间的距离，返回形状为 (n_points, len(sampled_indices)) 的矩阵
        distances_to_samples = cdist(points, selected_data)
        min_distances = np.min(distances_to_samples, axis=1)  # 每个点到现有采样点集的最小距离

    else:
        # 如果没有初始点，则随机选择一个作为第一个点
        first_index = np.random.randint(n_points)
        sampled_indices.append(first_index)
        # 计算所有点到第一个点的距离
        min_distances = np.linalg.norm(points - points[first_index], axis=1)

    # 进行最远点采样
    while len(sampled_indices) < n_samples:
        # 找到距离采样点集最远的点
        current_index = np.argmax(min_distances)

        # 如果没有点能满足最小距离要求，则提前终止
        if min_distances[current_index] < min_dist:
            break

        # 添加当前点到采样集
        sampled_indices.append(current_index)

        # 更新最小距离：仅计算当前点到新选择点的距离
        # 获取当前点到所有其他点的距离
        new_point = points[current_index]
        new_distances = np.linalg.norm(points - new_point, axis=1)

        # 更新每个点到现有样本点集的最小距离
        min_distances = np.minimum(min_distances, new_distances)
    return sampled_indices


import numpy as np



def select_structures(train, new_atoms ,descriptor, max_selected=20, min_distance=0.01 ):
    # 首先去掉跑崩溃的结构


    descriptor_function=get_descriptor_function(descriptor)

    train_des = np.array([np.mean(descriptor_function(i, descriptor), axis=0) for i in train])

    new_des = np.array([np.mean(descriptor_function(i, descriptor), axis=0) for i in new_atoms])

    selected_i =farthest_point_sampling(new_des,max_selected,min_distance,selected_data=train_des)

    return [new_atoms[i  ] for i in selected_i]

# 加速计算每对元素的最小键长
def compute_min_bond_lengths(atoms ):
    # 获取原子符号
    dist_matrix = atoms.get_all_distances()

    symbols = atoms.get_chemical_symbols()

    # 提取上三角矩阵（排除对角线）
    i, j = np.triu_indices(len(atoms), k=1)

    # 用字典来存储每种元素对的最小键长
    bond_lengths = {}

    # 遍历所有原子对，计算每一对元素的最小键长
    for idx in range(len(i)):
        atom_i, atom_j = symbols[i[idx]], symbols[j[idx]]
        # if atom_i==atom_j:
        #     continue
        # 获取当前键长
        bond_length = dist_matrix[i[idx], j[idx]]
        # if bond_length>5:
        #     continue
        # 确保元素对按字母顺序排列，避免 Cs-Ag 和 Ag-Cs 视为不同
        element_pair = tuple(sorted([atom_i, atom_j]))

        # 如果该元素对尚未存在于字典中，初始化其最小键长
        if element_pair not in bond_lengths:
            bond_lengths[element_pair] = bond_length
        else:
            # 更新最小键长
            bond_lengths[element_pair] = min(bond_lengths[element_pair], bond_length)

    return bond_lengths



def process_trajectory(trajectory):
    # 读取轨迹文件（假设为 xyz 格式）


    # 存储所有帧的结果
    all_bond_lengths = []

    # 遍历轨迹中的每一帧
    for atoms in trajectory:
        # 获取当前帧的距离矩阵

        # 计算当前帧的最小键长
        bond_lengths = compute_min_bond_lengths(atoms )
        all_bond_lengths.append(bond_lengths)

    return all_bond_lengths

def filter_by_bonds(trajectory,model):
    good_structure=[]
    bad_structure=[]
    base_bond=compute_min_bond_lengths(model )
    bonds=process_trajectory(trajectory)
    for index,bond in enumerate(bonds):
        # print(bond)
        # condition = [utils.radius_table[a]+utils.radius_table[b] > a_b for (a,b),a_b in bond.items()]

        condition = [base_bond.get(key,0)*0.6 > a_b for key,a_b in bond.items()]
        # print(condition)
        if any(condition):
            bad_structure.append(trajectory[index])
        else:
            good_structure.append(trajectory[index])
    return good_structure, bad_structure


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/28 21:38
# @Author  : 兵
# @email    : 1747193328@qq.com
#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024/6/21 16:40
# @Author  : 兵
# @email    : 1747193328@qq.com
import glob
import os

import matplotlib

from NepTrain import utils

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import numpy as np
try:
    from sklearn.metrics import r2_score, mean_squared_error
except ImportError:
    r2_score=None
    mean_squared_error=None
Config = [
    {"name": "energy", "unit": "eV/atom"},
    {"name": "force", "unit": "eV/A"},
    {"name": "virial", "unit": "eV/atom"},
    {"name": "stress", "unit": "GPa"},
]


def plot_loss_result(axes: plt.Axes):
    loss = np.loadtxt("loss.out")
    axes.loglog(loss[:, 1:7],
                label=['Total', 'L1-regularization',
                       'L2-regularization', 'Energy-train',
                       'Force-train', 'Virial-train'])
    axes.set_xlabel('Generation/100')
    axes.set_ylabel('Loss')
    if np.any(loss[7:10] != 0):
        axes.loglog(loss[:, 7:10], label=['Energy-test', 'Force-test', 'Virial-test'])
    axes.legend(ncol=2, frameon=False)


def plot_train_result(axes: plt.Axes, config: dict):
    types = ["train", "test"]
    colors = ['deepskyblue', 'orange']
    xys = [(0.1, 0.7), (0.4, 0.1)]
    for i in range(2):
        data_type = types[i]
        color = colors[i]
        xy = xys[i]
        if not os.path.exists(f"{config['name']}_{data_type}.out"):
            continue
        data = np.loadtxt(f"{config['name']}_{data_type}.out")
        min_value = np.min(data)
        max_value = np.max(data)
        index = data.shape[1] // 2
        axes.plot(data[:, index:], data[:, :index], '.', color=color, label=data_type)
        axes.plot(np.linspace(min_value, max_value, num=10), np.linspace(min_value, max_value, num=10), '-', color="k")

        if mean_squared_error is not None:
            rmse = np.sqrt(mean_squared_error(data[:, :index], data[:, index:]))
            r2 = r2_score(data[:, :index], data[:, index:])
            axes.text(xy[0], xy[1],
                      f'{data_type} RMSE={1000 * rmse:.3f}({"m" + config["unit"] if config["name"] != "stress" else "MPa"} )\n{data_type} $R^2$={r2:.3f}',
                      transform=axes.transAxes, fontsize=13)
        else:
            axes.text(xy[0], xy[1],
                      f'{data_type}',
                      transform=axes.transAxes, fontsize=13)
    handles, labels = axes.get_legend_handles_labels()
    label_dict = dict(zip(labels, handles))
    axes.legend(label_dict.values(), label_dict, frameon=False, ncol=2, columnspacing=1)
    axes.set_xlabel(f'DFT {config["name"]} ({config["unit"]})')
    axes.set_ylabel(f'NEP {config["name"]} ({config["unit"]})')


def _plot_nep_result( ):
    out_num = len(glob.glob("*_train.out"))
    if os.path.exists("loss.out"):
        out_num+=1
    test_out_num = len(glob.glob("*test.out"))

    rows = 2 if out_num >= 4 else 1
    cols = (out_num - test_out_num) // rows + (out_num - test_out_num) % rows


    fig = plt.figure(figsize=(6 * cols, 5 * rows))
    grids = fig.add_gridspec(rows, cols)

    if os.path.exists("loss.out"):
        axes_index = 0
        axes = fig.add_subplot(grids[axes_index])
        axes_index += 1

        plot_loss_result(axes)
    else:
        axes_index = 0

    for config in Config:
        if not os.path.exists(f"{config['name']}_train.out"):
            continue
        axes = fig.add_subplot(grids[axes_index])
        plot_train_result(axes, config)
        axes_index += 1

    plt.subplots_adjust(left=0.1, right=0.95, bottom=0.1, top=0.95)
    plt.savefig("nep_result.png", dpi=300)


def plot_nep_result(directory):
    with utils.cd(directory):
        _plot_nep_result()



if __name__ == '__main__':
    plot_nep_result("./")


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 15:53
# @Author  : 兵
# @email    : 1747193328@qq.com
import os
import subprocess

from ase.calculators.vasp import Vasp
from ase.io import read as ase_read

from NepTrain import Config


class VaspInput(Vasp):


    def __init__(self,*args,**kwargs):

        super(VaspInput,self).__init__(*args,**kwargs)
        self.input_params["setups"] = {"base": "recommended"}
        self.input_params["pp"] = ''
        os.environ[self.VASP_PP_PATH] = os.path.expanduser(Config.get("environ", "potcar_path"))


        # self.converged
    def _run(self, command=None, out=None, directory=None):
        """Method to explicitly execute VASP"""
        if command is None:
            command = self.command
        if directory is None:
            directory = self.directory

        errorcode = subprocess.call(command,
                                    shell=True,
                                    stdout=out,
                                    stderr=out,
                                    cwd=directory)

        return errorcode
if __name__ == '__main__':
    vasp=VaspInput()




    atoms=ase_read("./POSCAR",format='vasp')
    vasp.read_incar("./INCAR")
    vasp.calculate(atoms,('energy'))
    print(vasp.results)
    print(vasp.atoms.info)
    print(atoms.calc.results)


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/13 19:36
# @Author  : 兵
# @email    : 1747193328@qq.com
import os.path

from NepTrain import utils

from ase.io import read as ase_read
from ase.io import write as ase_write
from .select import select_structures, filter_by_bonds
from ..gpumd.plot import plot_md_selected
from ..nep.utils import read_symbols_from_file
from dscribe.descriptors import SOAP

def run_select(argparse):

    if utils.is_file_empty(argparse.trajectory_path):
        raise FileNotFoundError(f"An invalid file path was provided: {argparse.trajectory_path}.")
    utils.print_msg("Reading the file, please wait...")

    trajectory=ase_read(argparse.trajectory_path,":",format="extxyz")

    # if argparse.filter:
    # #转移到gpumd的模块
    #     atoms=ase_read(argparse.base_structure,)
    #     good, bad = filter_by_bonds(trajectory, model=atoms)
    #     directory=os.path.dirname(argparse.trajectory_path)
    #     trajectory=good
    #     ase_write(os.path.join(directory, "good_structures.xyz"), good,append=False)
    #     ase_write(os.path.join(directory, "remove_by_bond_structures.xyz"), bad,append=False)
    #     utils.print_msg(f"Bond length filtering activated, {len(bad)} structures filtered out, saved to {os.path.join(directory, 'remove_by_bond_structures.xyz')}.")


    if utils.is_file_empty(argparse.base):
        base_train=[]
    else:
        base_train=ase_read(argparse.base,":",format="extxyz")
    if utils.is_file_empty(argparse.nep):
        utils.print_msg("An invalid path for nep.txt was provided, using SOAP descriptors instead.")
        species=set()
        for atoms in trajectory+base_train:
            for i in atoms.get_chemical_symbols():
                species.add(i)

        species = list(species)
        r_cut = argparse.r_cut
        n_max = argparse.n_max
        l_max = argparse.l_max
        descriptor = SOAP(
            species=species,
            periodic=False,
            r_cut=r_cut,
            n_max=n_max,
            l_max=l_max,
        )

    else:
        descriptor=argparse.nep
    utils.print_msg("Starting to select points, please wait...")

    selected_structures = select_structures(base_train,
                                            trajectory,
                                            descriptor,
                      max_selected=argparse.max_selected,
                      min_distance=argparse.min_distance,
                      )

    utils.print_msg(f"Obtained {len(selected_structures)} structures." )

    ase_write(argparse.out_file_path, selected_structures)
    png_path=os.path.join(os.path.dirname(argparse.out_file_path),"selected.png")
    plot_md_selected(argparse.base,
                     argparse.trajectory_path,

                     argparse.out_file_path,
                     descriptor,
                       png_path ,

                     )
    utils.print_msg(f"The point selection distribution chart is saved to {png_path}." )
    utils.print_msg(f"The selected structures are saved to {argparse.out_file_path}." )




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 15:06
# @Author  : 兵
# @email    : 1747193328@qq.com

import os.path

from ase import Atoms
from ase.io import read as ase_read
from ase.io import write as ase_write

from NepTrain import utils
from ..select import select_structures, filter_by_bonds
from .io import RunInput
from .plot import plot_md_selected
from ..utils import check_env

atoms_index = 0

@utils.iter_path_to_atoms(["*.vasp","*.xyz"],show_progress=False)
def calculate_gpumd(atoms:Atoms,argparse):
    global atoms_index
    atoms_index+=1

    new_atoms=[]


    for temperature in argparse.temperature:

        run = RunInput(argparse.nep_txt_path)
        run.read_run(argparse.run_in_path)
        run.set_time_temp(argparse.time,temperature)
        directory=os.path.join(argparse.directory,f"{atoms_index}-{atoms.symbols}@{temperature}k-{argparse.time}ps")
        utils.print_msg(f"GPUMD is running, temperature: {temperature}k. Time: {argparse.time*run.time_step}ps" )

        run.calculate(atoms,directory)

        dump = ase_read(os.path.join(directory,"dump.xyz"), ":", format="extxyz", do_not_split_by_at_sign=True)
        for i, atom in enumerate(dump):
            atom.info["Config_type"] = f"{atom.symbols}-epoch-{argparse.time}ps-{temperature}k-{i + 1}"


        if argparse.filter:
            good,bad=filter_by_bonds(dump,model=atoms)
            dump=good
            ase_write(os.path.join(directory,"remove_by_bond_structures.xyz"),bad)
        ase_write(argparse.out_file_path,dump,append=True)







    return new_atoms
def run_gpumd(argparse):
    check_env()
    utils.verify_path(os.path.dirname(argparse.out_file_path))
    result = calculate_gpumd(argparse.model_path,argparse)












#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 21:52
# @Author  : 兵
# @email    : 1747193328@qq.com
import os

from NepTrain import Config
from NepTrain import utils


def check_env():
    if not os.path.exists(os.path.expanduser(Config.get("environ", "potcar_path"))):
        raise FileNotFoundError("Please edit the pseudopotential file path in ~/.NepTrain to set a valid path!")

    for option in ["vasp_path", "mpirun_path", "nep_path", "gpumd_path"]:
        if utils.get_command_result(["which", Config.get("environ", option)]) is None:
            utils.print_warning(f"The environment variable {option.replace('_path', '')} is not set. If you have set the environment in the submission script, please ignore this warning.")




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 15:20
# @Author  : 兵
# @email    : 1747193328@qq.com
from .io import RunInput,PredictionRunInput
from .run import run_nep
from .plot import plot_nep_result


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/27 13:26
# @Author  : 兵
# @email    : 1747193328@qq.com
"""
复制的hiphive\structure_generation\rattle.py
因为这个包依赖太多了 而且目前只用到了这个微扰  所以先只复制过来
"""
import numpy as np
from ase.neighborlist import NeighborList
from scipy.special import erf


def _probability_mc_rattle(d, d_min, width):
    """ Monte Carlo probability function as an error function.

    Parameters
    ----------
    d_min : float
        center value for the error function
    width : float
        width of error function
    """

    return (erf((d-d_min)/width) + 1.0) / 2
def mc_rattle(atoms, rattle_std, d_min, width=0.1, n_iter=10, max_attempts=5000,
              max_disp=2.0, active_atoms=None, nbr_cutoff=None, seed=42):
    """Generate displacements using the Monte Carlo rattle method

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    rattle_std : float
        rattle amplitude (standard deviation in normal distribution)
    d_min : float
        interatomic distance used for computing the probability for each rattle
        move. Center position of the error function
    width : float
        width of the error function
    n_iter : int
        number of Monte Carlo cycle
    max_disp : float
        rattle moves that yields a displacement larger than max_disp will
        always be rejected. This rarley occurs and is more used as a safety net
        for not generating structures where two or more have swapped positions.
    max_attempts : int
        limit for how many attempted rattle moves are allowed a single atom;
        if this limit is reached an `Exception` is raised.
    active_atoms : list
        list of which atomic indices should undergo Monte Carlo rattling
    nbr_cutoff : float
        The cutoff used to construct the neighborlist used for checking
        interatomic distances, defaults to 2 * d_min
    seed : int
        seed for setting up NumPy random state from which random numbers are
        generated

    Returns
    -------
    numpy.ndarray
        atomic displacements (`Nx3`)
    """

    # setup
    rs = np.random.RandomState(seed)

    if nbr_cutoff is None:
        nbr_cutoff = 2 * d_min

    if active_atoms is None:
        active_atoms = range(len(atoms))

    atoms_rattle = atoms.copy()
    reference_positions = atoms_rattle.get_positions()
    nbr_list = NeighborList([nbr_cutoff/2]*len(atoms_rattle), skin=0.0,
                            self_interaction=False, bothways=True)
    nbr_list.update(atoms_rattle)

    # run Monte Carlo
    for _ in range(n_iter):
        for i in active_atoms:
            i_nbrs = np.setdiff1d(nbr_list.get_neighbors(i)[0], [i])
            for n in range(max_attempts):

                # generate displacement
                delta_disp = rs.normal(0.0, rattle_std, 3)
                atoms_rattle.positions[i] += delta_disp
                disp_i = atoms_rattle.positions[i] - reference_positions[i]

                # if total displacement of atom is greater than max_disp, then reject delta_disp
                if np.linalg.norm(disp_i) > max_disp:
                    # revert delta_disp
                    atoms_rattle[i].position -= delta_disp
                    continue

                # compute min distance
                if len(i_nbrs) == 0:
                    min_distance = np.inf
                else:
                    min_distance = np.min(atoms_rattle.get_distances(i, i_nbrs, mic=True))

                # accept or reject delta_disp
                if _probability_mc_rattle(min_distance, d_min, width) > rs.rand():
                    # accept delta_disp
                    break
                else:
                    # revert delta_disp
                    atoms_rattle[i].position -= delta_disp
            else:
                raise Exception(f'Exceeded the maximum number of attempts for atom {i}')
    displacements = atoms_rattle.positions - reference_positions
    return displacements
def get_cell_perturb_matrix(cell_pert_fraction: float):
    if cell_pert_fraction < 0:
        raise RuntimeError("cell_pert_fraction can not be negative")
    e0 = np.random.rand(6)
    e = e0 * 2 * cell_pert_fraction - cell_pert_fraction
    cell_pert_matrix = np.array(
        [
            [1 + e[0], 0.5 * e[5], 0.5 * e[4]],
            [0.5 * e[5], 1 + e[1], 0.5 * e[3]],
            [0.5 * e[4], 0.5 * e[3], 1 + e[2]],
        ]
    )
    return cell_pert_matrix

def generate_mc_rattled_structures(atoms, n_structures,cell_pert_fraction, rattle_std, d_min,
                                   seed=42, **kwargs):
    """Returns list of Monte Carlo rattled configurations.

    Rattling atom `i` is carried out as a Monte Carlo move that is
    accepted with a probability determined from the minimum
    interatomic distance :math:`d_{ij}`.  If :math:`\\min(d_{ij})` is
    smaller than :math:`d_{min}` the move is only accepted with a low
    probability.

    This process is repeated for each atom a number of times meaning
    the magnitude of the final displacements is not *directly*
    connected to `rattle_std`.

    Warning
    -------
    Repeatedly calling this function *without* providing different
    seeds will yield identical or correlated results. To avoid this
    behavior it is recommended to specify a different seed for each
    call to this function.

    Notes
    ------
    The procedure implemented here might not generate a symmetric
    distribution for the displacements `kwargs` will be forwarded to
    `mc_rattle` (see user guide for a detailed explanation)

    The displacements generated will roughly be n_iter**0.5 * rattle_std
    for small values of n_iter

    Parameters
    ----------
    atoms : ase.Atoms
        prototype structure
    n_structures : int
        number of structures to generate
    rattle_std : float
        rattle amplitude (standard deviation in normal distribution);
        note this value is not connected to the final
        average displacement for the structures
    d_min : float
        interatomic distance used for computing the probability for each rattle
        move
    seed : int
        seed for setting up NumPy random state from which random numbers are
        generated
    n_iter : int
        number of Monte Carlo cycles (iterations), larger number of iterations will
        generate larger displacements (defaults to 10)

    Returns
    -------
    list of ase.Atoms
        generated structures
    """
    rs = np.random.RandomState(seed)
    atoms_list = []

    for _ in range(n_structures):
        atoms_tmp = atoms.copy()
        seed = rs.randint(1, 1000000000)


        cell_perturb_matrix = get_cell_perturb_matrix(cell_pert_fraction)

        cell = np.matmul(
            atoms_tmp.cell.array, cell_perturb_matrix
        )
        atoms_tmp.set_cell(cell,True)
        displacements = mc_rattle(atoms_tmp, rattle_std, d_min, seed=seed, **kwargs)
        atoms_tmp.positions += displacements
        atoms_list.append(atoms_tmp)
    return atoms_list




#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/25 19:10
# @Author  : 兵
# @email    : 1747193328@qq.com
from ase.calculators.vasp.setups import setups_defaults

from NepTrain import Config
from .io import VaspInput
from .run import run_vasp

for option in Config.options("potcar"):
    v=Config.get("potcar", option).replace(option, "")

    setups_defaults["recommended"][option.title()]=v





#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/24 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
"""Pymatgen package configuration."""

from __future__ import annotations

import platform
import sys

import numpy as np
from setuptools import Extension, setup

is_win_64 = sys.platform.startswith("win") and platform.machine().endswith("64")
extra_link_args = ["-Wl,--allow-multiple-definition"] if is_win_64 else []

setup(
    author="Chen Cheng bing",

    include_dirs=[np.get_include()],
)


#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/10/29 19:56
# @Author  : 兵
# @email    : 1747193328@qq.com
import os.path

import numpy as np
from ase.io import read as ase_read
from calorine.gpumd import read_thermo


from matplotlib import pyplot as plt
from sklearn.decomposition import PCA
from NepTrain import utils
from NepTrain.core.nep.utils import get_descriptor_function


def plot_md_selected(train_xyz_path,md_xyz_path,selected_xyz_path,descriptor,save_path):
    # 画一下图
    config = [
        # (文件名,图例,图例颜色)

    ]
    if not utils.is_file_empty(train_xyz_path):
        config.append((train_xyz_path, "base dataset","gray"))

    if not utils.is_file_empty(md_xyz_path):
        config.append((md_xyz_path, 'new dataset', "#07cd66"))

    if not utils.is_file_empty(selected_xyz_path):
        config.append((selected_xyz_path,'selected', "red"))

    fit_data = []

    descriptor_function=get_descriptor_function(descriptor)


    for info in config:
        atoms_list = ase_read(info[0], ":", format="extxyz", do_not_split_by_at_sign=True)

        # atoms_list_des = np.vstack([get_descriptors(i, nep_txt_path) for i in atoms_list])
        atoms_list_des = np.array([np.mean(descriptor_function(i, descriptor), axis=0) for i in atoms_list])


        fit_data.append(atoms_list_des)

    reducer = PCA(n_components=2)
    reducer.fit(np.vstack(fit_data))
    fig = plt.figure()
    for index, array in enumerate(fit_data):
        proj = reducer.transform(array)
        plt.scatter(proj[:, 0], proj[:, 1], label=config[index][1], c=config[index][2])

    plt.legend()
    # plt.axis('off')
    plt.savefig(save_path)
    plt.close(fig)






def plot_all_structure(train_data, add_data,descriptor, save_path):
    descriptor_function=get_descriptor_function(descriptor)

    train_des = np.array([np.mean(descriptor_function(i, descriptor), axis=0) for i in train_data])
    add_des = np.array([np.mean(descriptor_function(i, descriptor), axis=0) for i in add_data])

    reducer = PCA(n_components=2)
    reducer.fit(np.vstack([train_des, add_des]))

    fig = plt.figure()

    proj = reducer.transform(train_des)
    plt.scatter(proj[:, 0], proj[:, 1], label='train', c="gray")

    proj = reducer.transform(add_des)
    plt.scatter(proj[:, 0], proj[:, 1], label='add', c="#07cd66")

    plt.legend()
    # plt.axis('off')

    plt.savefig(save_path)
    plt.close(fig)


def plot_energy(thermo_path,natoms=1):
    data = read_thermo(thermo_path, natoms)

    potential_energy = data.potential_energy.to_numpy(dtype='float')

    fig = plt.figure()
    plt.plot(list(range(potential_energy.shape[0])), potential_energy)



    plt.savefig(os.path.join(os.path.dirname(thermo_path),"md_energy.png"), dpi=300)
    plt.close(fig)
